// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: http://codemirror.net/LICENSE

(function(mod) {
  if (typeof exports == "object" && typeof module == "object") // CommonJS
    mod(require("../../lib/codemirror"));
  else if (typeof define == "function" && define.amd) // AMD
    define(["../../lib/codemirror"], mod);
  else // Plain browser env
    mod(CodeMirror);
})(function(CodeMirror) {
  "use strict";

  CodeMirror.defineMode("sqf", function(config, parserConfig) {
    const ST_DEFINITION       = Symbol("state definition");
    const ST_EXPRESSION       = Symbol("state expression");
    const ST_RENAME_PARAM     = Symbol("state rename param");
    const ST_SCHEMA           = Symbol("state schema");
    const ST_SELECT_CONDITION = Symbol("state select condition");
    const ST_TUPLESET         = Symbol("state tuple set");
    const ST_TUPLE            = Symbol("state tuple");
    
    
    function words(str) {
      var obj = {}, words = str.split(" ");
      for (var i = 0; i < words.length; ++i) 
        obj[words[i]] = true;
      return obj;
    }
    
    let keywords = words(
      "@let @relation @select @project @join @rename @union @diff @s @p @r"
    );
    
    let regex = {
      id: /[a-zA-Z][a-zA-Z1-9âêîôûäëïöüéèàùçÂÊÎÔÛÄËÏÖÜÉÈÀÙÇ_]*/,
    };
    
    return {
      startState: function() {
        return {
          current: null,
          lastKeyword: null,
        }
      },
      
      
      token: function(stream, state) {
        const regexId = /[a-zA-ZâêîôûäëïöüéèàùçÂÊÎÔÛÄËÏÖÜÉÈÀÙÇ][a-zA-Z0-9âêîôûäëïöüéèàùçÂÊÎÔÛÄËÏÖÜÉÈÀÙÇ_#]*/;
        const regexDquoted = /\"([^"]|\\\")*\"/;
        const regexSquoted = /\'([^']|\\\')*\"/;
        const regexUnquoted = /([^,>]|\\,|\\>)+(,|>)/;
        
        // COMMON TO ALL STATES
        
        // ignore whitespaces
        if (stream.eatSpace())
          return;
        
        console.assert(stream.peek() != " ");

        // comment -> skip to end of line
        if (stream.peek() === "%") {
          stream.skipToEnd();
          return "comment";
        }
        
        if (stream.match("@let")) {
          state.current = ST_DEFINITION;
          return "keyword";
        }
        
        if (stream.match("@print")) {
          state.current = ST_EXPRESSION;
          return "keyword";
        }
        
        // STATE DEFINITION
        if (state.current == ST_DEFINITION) {
          
          if (stream.match(regexId)) {
            return "variable";
          }
          
          if (stream.match("=")) {
            state.current = ST_EXPRESSION;
            return;
          }
        }

        // STATE EXPRESSION
        if (state.current == ST_EXPRESSION) {
          let token;
          if (token = stream.match(/@[A-Za-z]*/)) {
            if (! keywords[token[0]]) {
              state.lastKeyword = null;
              return "error";
            }
            else {
              state.lastKeyword = token[0];
              return "keyword";
            }
          }
          
          if (token = stream.match(/[\+\-\*]/)) {
              state.lastKeyword = null;
              return "keyword";
          }
          
          if (stream.match(regexId)) {
            return "variable";
          }
          
          if ( stream.match(/[\(\)]/) ) {
            return;
          }
          
          if (stream.eat('{')) {
            if (state.lastKeyword == "@relation") {
              state.current = ST_TUPLESET;
              return;
            }
            
            if (state.lastKeyword == "@select" || state.lastKeyword == "@s" ) {
              state.current = ST_SELECT_CONDITION;
              return;
            }
            
            if (state.lastKeyword == "@project" || state.lastKeyword == "@p" ) {
              state.current = ST_SCHEMA;
              return;
            }
            
            if (state.lastKeyword == "@rename" || state.lastKeyword == "@r" ) {
              state.current = ST_RENAME_PARAM;
              return;
            }
            
            return "error";
          }
          
          if (stream.eat('[') && state.lastKeyword == "@relation") {
              state.current = ST_SCHEMA;
              return "bracket";
          }
        }
        
        // STATE SCHEMA
        if (state.current == ST_SCHEMA) {
          if (stream.match(regexId)) {
            return "def";
          }
          if (stream.eat(",")) {
            return;
          }
          if (stream.eat("]") && state.lastKeyword == "@relation") {
            state.current = ST_EXPRESSION;
            return;
          }
          if (stream.eat("}") && ( state.lastKeyword == "@project" || state.lastKeyword == "@p" ) ) {
            state.current = ST_EXPRESSION;
            state.lastKeyword = null;
            return;
          }
        }

        // STATE SELECT_CONDITION
        if (state.current == ST_SELECT_CONDITION) {
          console.log("état SELECT_CONDITION");
          if (stream.match(regexId)) {
            return "attribute";
          }
          
          if (stream.match(/=|<>|<=|>=|<|>/)) {
            console.log("opérateur de sélection ???");
            return "operator";
          }
          if (stream.match(regexDquoted) || stream.match(regexSquoted)) {
            return "string";
          }
          if (stream.eat("}")) {
            state.lastKeyword = null;
            state.current = ST_EXPRESSION;
            return;
          }
        }
        
        // STATE RENAME_PARAM
        if (state.current == ST_RENAME_PARAM) {
          if (stream.match(regexId)) {
            return "attribute";
          }
          if (stream.eat(":")) {
            return;
          }
          if (stream.eat("}")) {
            state.lastKeyword = null;
            state.current = ST_EXPRESSION;
            return;
          }
        }
        
        // STATE TUPLE SET
        if (state.current == ST_TUPLESET) {
          if (stream.eat("<")) {
            state.current = ST_TUPLE;
            return;
          }
          if (stream.eat("}")) {
            state.lastKeyword = null;
            state.current = ST_EXPRESSION;
            return;
          }
        }
        
        // STATE TUPLE
        if (state.current == ST_TUPLE) {
          if (stream.match(regexDquoted) 
                || stream.match(regexSquoted) ) {
            return "string";
          }
          let str;
          if (str = stream.match(regexUnquoted)) {
            stream.backUp(1);
            return "string";
          }
          if (stream.eat(",")) {
            return;
          }
          if (stream.eat(">")) {
            state.current = ST_TUPLESET;
            return;
          }
        }
        
        stream.skipToEnd();
        console.log("erreur sur", `[${stream.current()}]`, stream.current().length );
        // for (let c of stream.current())
          // console.log("char : ", c.charCodeAt(0));
        console.log(stream);
        return "error";
        
      }
    };
  });

  // TODO : usage ???
  CodeMirror.defineMIME("text/sqf", "sqf", {
    keywords: "@relation",
  });

});
