'use strict'

class Rename extends Expression {
  
  // map is an object for which each property is the old attribute name and each property value
  // is a the new attribute name.
  // for instance : {a1: "b1", a2: "b2"} stands for 
  //  - attribute "a1" will be renamed to "b1"
  //  - attribute "a2" will be renamed to "b2"
  constructor({
        expression=null, 
        map={} 
      } = {} ) {
        
    if (ForbiddenOperations.instance.isForbidden("R"))
      throw new AlgebraError(AlgebraError.FORBIDDEN_OPERATION, "Rename");
        
    super();
    this._expression = expression;
    this._parameters = Immutable.Map(map).filter(value => typeof value == "string");
    
    this._check();
    
    this._computeSchema();
    
    let mapString = Immutable.Map(map).map((v,k) => k + ":" + v).join(", ");
    
    super._expressionString = `@rename{${mapString}}${expression.toExpressionString(this)}`;
  }
  
  get level() {
    return 0;
  }
  
  get paramCount() {
    return this._parameters.size;
  }
  
  _evaluate() {
    let tuples = Immutable.Seq(this._expression.tupleSet).map(tuple => new Tuple(tuple)).toArray();
    
    delete this._expression;
    
    this._setTupleSet(new TupleSet(tuples));
  }
  
  _computeSchema() {
    let operandSchema = this._expression.schema;
    let attribs = [];
    
    for (let attrib of operandSchema) {
      let newAttrib = this._parameters.find(
        (value, key) => key == attrib,
        null,       // context ???
        attrib
      );
      
      attribs.push(newAttrib);
    }
    
    this._setSchema(new Schema(attribs));
  }
  
  _check() {
    if ( this._expression == null || ! (this._expression instanceof Expression) )
      throw new AlgebraError(AlgebraError.INVALID_EXPRESSION, this._expression);
    
    if ( ! (this._parameters instanceof Object) )
      throw new AlgebraError(AlgebraError.INVALID_RENAME_MAP, this._parameters);
    
    for ( let [oldName,newName] of this._parameters.entries() ) {
      if ( ! this._expression.schema.hasAttribute(oldName) )
        throw new AlgebraError(
          AlgebraError.INVALID_RENAME_MAP, 
          `cannot rename attribute '${oldName}' to '${newName}' : '${oldName}' is not present in schema ${this._expression.schema.toString()}`);
      
      if ( this._expression.schema.hasAttribute(newName)
            && ! this._parameters.has(newName) )
        throw new AlgebraError(
          AlgebraError.INVALID_RENAME_MAP, 
          `cannot rename attribute '${oldName}' to '${newName}' : '${newName}' is already present in schema ${this._expression.schema.toString()}`);
    }
    
    this._checkTargetNamesUnique();
  }
  
  _checkTargetNamesUnique() {
    // let seq = Immutable.Seq.of(this._parameters).filter(value => typeof value == "string").sort();
    let seq = Immutable.Seq(this._parameters).sort();
    
    let last = null;
    for (let cur of seq.values()) {
      if (cur == last)
        throw new AlgebraError(
          AlgebraError.INVALID_RENAME_MAP, 
          `attribute '${cur}' is used more than once in mapping ${this._parameters}`);
      last = cur;
    }
    
    return true;
  }
}