'use strict'

class Schema {
  
  constructor(...args) {
    let set = Immutable.Seq();
    for (let arg of args){
      set = Schema._addToSet(set, arg);
    }
    
    Schema._checkNoDuplicate(set);
    
    this._attributes = set.toIndexedSeq();
    this._checkNoEmptyAttributeNames();
  }
  
  get attributes() {
    return this._attributes;
  }
  
  get size() {
    return this._attributes.size;
  }
  
  getAttributeIndex(attrib){
    return this._attributes.keyOf(attrib);
  }
  
  getAttributeAt(index){
    if (typeof(index) !== "number" || index%1 !== 0)
      throw new AlgebraError(AlgebraError.INVALID_INDEX, `${index} is not an integer`);
    if (index <= -1 || index >= this.size)
      throw new AlgebraError(AlgebraError.INVALID_INDEX, `${index} is out of bounds ([${0}..${this.size}[)`);
    return this._attributes.get(index, undefined);
  }
  
  hasAttribute(attrib){
    return this._attributes.includes(attrib);
  }
  
  getUnion(other){
    if (! other instanceof Schema)
      throw new AlgebraError(AlgebraError.INVALID_SCHEMA);
    
    let setLeft = Immutable.OrderedSet(this._attributes);
    let setRight = Immutable.OrderedSet(other._attributes);
    let union = setLeft.union(setRight);
    
    return new Schema(...union.values());
  }
  
  getIntersection(other){
    if (! other instanceof Schema)
      throw new AlgebraError(AlgebraError.INVALID_SCHEMA);
    
    let setLeft = Immutable.OrderedSet(this._attributes);
    let setRight = Immutable.OrderedSet(other._attributes);
    let intersection = setLeft.intersect(setRight);
    
    return new Schema(...intersection.values());
  }
  
  getDifference(other){
    if (! other instanceof Schema)
      throw new AlgebraError(AlgebraError.INVALID_SCHEMA);
    
    let setLeft = Immutable.OrderedSet(this._attributes);
    let setRight = Immutable.OrderedSet(other._attributes);
    let diff = setLeft.subtract(setRight);
    
    return new Schema(...diff.values());
  }
  
  getSymmetricDifference(other){
    if (! other instanceof Schema)
      throw new AlgebraError(AlgebraError.INVALID_SCHEMA);
    
    let setLeft = Immutable.OrderedSet(this._attributes);
    let setRight = Immutable.OrderedSet(other._attributes);
    let res = setLeft.subtract(setRight).union(setRight.subtract(setLeft));
    
    return new Schema(...res.values());
  }
  
  equals(other){
    if (! other instanceof Schema)
      throw new AlgebraError(AlgebraError.INVALID_SCHEMA);
    
    if (this === other)
      return true;
    
    if (this.length !== other.length)
      return false;
    
    return this.getSymmetricDifference(other).size === 0;
  }
  
  equalsStrict(other){
    if (! other instanceof Schema)
      throw new AlgebraError(AlgebraError.INVALID_SCHEMA);
    
    if (this === other)
      return true;
    
    if (this.length !== other.length)
      return false;
    
    for (let [index, attrib] of this.attributes.entries()) {
      if (other.getAttributeAt(index) !== attrib)
        return false;
    }
    
    return true;
  }
  
  // return an array containing indexes of another schema attributes in this schema
  // ex : ["a1", "a2, "a3", "a4"].getAttributeIndexes("a4", "a2") ==> [3, 1]
  getAttributeIndexes(other) {
    let res = [];
    
    for (let attrib of other) {
      res.push(this.getAttributeIndex(attrib));
    }
    
    return res;
  }
  
  toString() {
    return `${this._attributes.join(", ")}`;
  }
  
  toArray() {
    return this._attributes.toArray();
  }
  
  get attribSeq() {
    return this._attributes;
  }
  
  
  _checkNoEmptyAttributeNames() {
    for (let [index, elem] of this._attributes.entries()) {
      if (elem === "")
        throw new AlgebraError(AlgebraError.INVALID_ATTRIBUTE, `Undefined attribute at index ${index}`);
    }
  }
  
  static _addToSet(set, arg){
    if (arg instanceof Schema) {
      set = set.concat(arg._attributes);
    }
    else if (typeof arg === "string") {
      set = set.concat(arg.trim());
    }
    else if (arg instanceof Array) {
      for(let elem of arg){
        set = Schema._addToSet(set, elem);
      }
    }
    else
      throw new AlgebraError(AlgebraError.INVALID_ATTRIBUTE, `argument '${arg}' (${typeof arg}) cannot be used for schema creation`);
    
    return set;
  }
  
  static _checkNoDuplicate(set) {
    let sorted = set.sort();
    
    let prev = null;
    for (let a of sorted.values()) {
      if (a == prev)
        throw new AlgebraError(AlgebraError.INVALID_ATTRIBUTE, `duplicate attribute '${a}'`);
      prev = a;
    }
  }
 
  
  *[Symbol.iterator](){
    for(let e of this._attributes.values()){
      yield e;
    }
  }
}