'use strict'



class SelectCondition{
  static _initConstants(){
    SelectCondition.OP_EQ = Symbol("EQ");
    SelectCondition.OP_NEQ = Symbol("NEQ");
    SelectCondition.OP_LT = Symbol("LT");
    SelectCondition.OP_LTE = Symbol("LTE");
    SelectCondition.OP_GT = Symbol("GT");
    SelectCondition.OP_GTE = Symbol("GTE");
  }
  
  constructor({
          left={attrib:null, constant:null}, 
          operator=SelectCondition.OP_EQ,
          right={attrib:null, constant:null}
        } = {}) {
          
    SelectCondition._checkOperand(left, "left operand");
    SelectCondition._checkOperand(right, "right operand");
    
    if (left.constant != null && right.constant != null)
      throw new AlgebraError(AlgebraError.INVALID_CONDITION, "at least one operand must be an attribute");
    
    SelectCondition._checkOperator(operator);
    
    this._left = left;
    this._right = right;
    this._operator = operator;
    
    this.matches = SelectCondition._createMatchMethod(left, right, operator);
  }
  
  checkSchemaCompatibility(schema) {
    if (this._left.attrib != null && !schema.hasAttribute(this._left.attrib))
      throw new AlgebraError(AlgebraError.INVALID_CONDITION, `Schema ${schema} doesn't contain attribute '${this._left.attrib}'`);
    
    if (this._right.attrib != null && !schema.hasAttribute(this._right.attrib))
      throw new AlgebraError(AlgebraError.INVALID_CONDITION, `Schema ${schema} doesn't contain attribute '${this._right.attrib}'`);
  }
  
  toString() {
    let left = this._left.constant
         ? `'${this._left.constant}'`
         : `${this._left.attrib}`;
    
    let right = this._right.constant
         ? `'${this._right.constant}'`
         : `${this._right.attrib}`;
         
    let operator;
    
    switch(this._operator) {
      case SelectCondition.OP_EQ:
        operator = "=";
        break;
      case SelectCondition.OP_NEQ:
        operator = "<>";
        break;
      case SelectCondition.OP_LT:
        operator = "<";
        break;
      case SelectCondition.OP_LTE:
        operator = "<=";
        break;
      case SelectCondition.OP_GT:
        operator = ">";
        break;
      case SelectCondition.OP_GTE:
        operator = ">=";
        break;
    }
    
    return `${left}${operator}${right}`;
  }
  
  static _checkOperand(op, name) {
    if (op.attrib == null && op.constant == null
          || op.attrib != null && op.constant != null)
      throw new AlgebraError(
          AlgebraError.INVALID_CONDITION, 
          `${name} : undefined attrib or constant value`);
      
    if (op.attrib != null && typeof op.attrib != "string")
      throw new AlgebraError(
          AlgebraError.INVALID_CONDITION, 
          `${name} : 'attrib' property must be a string`);
    
    if (op.constant != null && typeof op.constant != "string")
      throw new AlgebraError(
          AlgebraError.INVALID_CONDITION, 
          `${name} : 'constant' property must be a string`);
  }
  
  static _checkOperator(operator) {
    if (   operator != SelectCondition.OP_EQ
        && operator != SelectCondition.OP_NEQ
        && operator != SelectCondition.OP_LT
        && operator != SelectCondition.OP_LTE
        && operator != SelectCondition.OP_GT
        && operator != SelectCondition.OP_GTE )
      throw new AlgebraError(
          AlgebraError.INVALID_CONDITION, 
          `operator must be one of : OP_EQ, OP_NEQ, OP_LT, OP_LTE, OP_GT, OP_GTE instead of '${operator}'`);
  }
  

  static _createMatchMethod(left, right, operator) {
    if ( ! ExtendedOperators.instance.isAllowed(operator) )
      throw new AlgebraError(
        AlgebraError.FORBIDDEN_OPERATOR,
        operator.toString());
        // `operator ${operator.toString()} has not been allowed in current algebra`);
    
    return function(tuple) {
      let leftValue = left.attrib ? tuple.getValue(left.attrib) : left.constant;
      let rightValue = right.attrib ? tuple.getValue(right.attrib) : right.constant;
      
      switch( operator ) {
        case SelectCondition.OP_EQ:
          return leftValue == rightValue;
          break;
        case SelectCondition.OP_NEQ:
          return leftValue != rightValue;
          break;
        case SelectCondition.OP_LT:
          return leftValue < rightValue;
          break;
        case SelectCondition.OP_LTE:
          return leftValue <= rightValue;
          break;
        case SelectCondition.OP_GT:
          return leftValue > rightValue;
          break;
        case SelectCondition.OP_GTE:
          return leftValue >= rightValue;
          break;
      }
    };
  }
  
}


SelectCondition._initConstants();
