'use strict'

class Tuple {
  
  constructor(...args) {
    this._schema = null;
    this._cachedObject = null;

    let seq = Immutable.Seq();
    for (let arg of args)
      seq = Tuple._addValue(seq, arg);
    
    this._values = seq;
  }
  
  static _addValue(seq, val) {
    if (typeof val == "string")
      return seq.concat(val);
    
    if (val instanceof Array) {
      for (let e of val){
        seq = Tuple._addValue(seq, e);
      }
      return seq;
    }
    
    if (val instanceof Tuple) {
      return seq.concat(val._values);
    }
    
    
    throw new AlgebraError(
        AlgebraError.INVALID_ATTRIBUTE_VALUE, 
        `value ${val} (${typeof val}) cannot be used as attribute value`);
  }
  
  cloneReordered(indexes) {
    if (! (indexes instanceof Array))
      throw new AlgebraError(AlgebraError.INTERNAL, "indexes should be an array of integer");
    
    if (indexes.length != this.size)
      throw new AlgebraError(AlgebraError.INTERNAL, "There should be an index for each value");
    
    let orderedValues = [];
    
    for (let i=0; i<this.size; i++) {
      console.assert(indexes[i] >= 0);
      console.assert(indexes[i] < this.size);
      console.assert(orderedValues[indexes[i]] == undefined);
      
      orderedValues[indexes[i]] = this._values.get(i);
    }
    
    return new Tuple(orderedValues);
  }
  
  compareTo(other){
    if (other === null || ! other instanceof Tuple)
      throw new AlgebraError(AlgebraError.INVALID_TUPLE, "argument must be a tuple");
    
    if (this.size != other.size)
      throw new AlgebraError(AlgebraError.INVALID_TUPLE, "both tuples must have the same size");
    
    if (this._schema === null || other._schema === null || this._schema.equalsStrict(other._schema)) {
      for (let i=0; i<this.size; i++) {
        let leftValue = this._values.get(i);
        let rightValue = other._values.get(i);
        if (leftValue < rightValue)
          return -1;
        if (leftValue > rightValue)
          return 1;
      }
    }
    else {
      if (! this._schema.equals(other._schema))
        throw new AlgebraError(
            AlgebraError.INCOMPATIBLE_SCHEMA, 
            "cannot compare tuples with different schema (${this.schema.toString()} VS ${other.schema.toString()})");
      
      for (let attrib of this._schema.attributes.sort()) {
        let leftValue = this.getValue(attrib);
        let rightValue = other.getValue(attrib);
        
        if (leftValue < rightValue)
          return -1;
        if (leftValue > rightValue)
          return 1;
      }
    }
    
    return 0;
  }
  
  equals(other){
    if (other === null || ! other instanceof Tuple)
      throw new AlgebraError(AlgebraError.INVALID_TUPLE, "argument must be a tuple");
    
    if (Object.is(this, other))
      return true;
    
    if (this.size != other.size)
      return false;
    
    if (this._schema !== null && other._schema !== null 
        && ! this._schema.equals(other._schema))
      return false;
    
    return this.compareTo(other) === 0;
  }
  
  asObject() {
    if (this._cachedObject === null){
      this._cachedObject = {};
      if (this._schema !== null){
        for (let attrib of this._schema){
          this._cachedObject[attrib] = this.getValue(attrib);
        }
      }
      else{
        for (let i=0; i<this.size; i++)
          this._cachedObject[i] = this.getValueAt(i);
      }
    }
    
    return Object.assign({}, this._cachedObject);
  }
  
  get valueSeq() {
    return this._values;
  }
  
  getValue(attrib) {
    if (this._schema === null)
      throw new AlgebraError(AlgebraError.INVALID_SCHEMA, "Schema has not been set");
    
    let index = this._schema.getAttributeIndex(attrib);
    if (index === -1)
      throw new AlgebraError(
        AlgebraError.INVALID_ATTRIBUTE, 
        `unknown attribute '${attrib} in schema ${this.schema.toString()}` );
    
    return this.getValueAt(index);
  }
  
  getValueAt(index) {
    if (typeof(index) !== "number" || index%1 !== 0)
      throw new AlgebraError(AlgebraError.INVALID_INDEX, `${index} is not an integer`);
    
    if (index <= -1 || index >= this.size)
      throw new AlgebraError(AlgebraError.INVALID_INDEX, `${index} is out of bounds ([${0}..${this.size}[)`);
    
    return this._values.get(index, undefined);
  }
  
  get size() {
    return this._values.size;
    
  }
  
  get schema(){
    return this._schema;
  }
  
  set schema(value){
    if (this._schema !== null &&  ! Object.is(this._schema, value))
      throw new AlgebraError(
        AlgebraError.INVALID_SCHEMA, 
        `schema is already set to ${this._schema.toString()} ; cannot be changed to another schema instance`);
    
    if (! (value instanceof Schema) )
      throw new AlgebraError(AlgebraError.INVALID_SCHEMA, `${value} (${typeof value}) is not a Schema instance`);
    
    if (value.size !== this.size)
      throw new AlgebraError(
        AlgebraError.INVALID_SCHEMA, 
        `cannot set schema ${value.toString()} (size: ${value.size}) for this tuple (size: ${this.size})` );
        
   this._cachedObject = null;
   this._schema = value;
  }
  
  static _escape(s) {
    
		let hasDoubleQuote = s.indexOf('"') != -1;
		let hasSingleQuote = s.indexOf('\\') != -1;
		let hasBackslash = s.indexOf('\\') != -1;
		let hasGreater = s.indexOf('>') != -1;
		let hasComma = s.indexOf(',') != -1;
    let startsWithSpace = s.startsWith(' ');
    let endsWithSpace = s.endsWith(' ');
    let isEmpty = s.length == 0;
    
    let needsEscape = hasDoubleQuote
        || hasSingleQuote
		    || hasBackslash
		    || hasGreater
		    || hasComma
        || startsWithSpace
        || endsWithSpace
        || isEmpty;
        
    if (! needsEscape)
      return s;

    let quoteDelimiter = (! hasDoubleQuote || hasSingleQuote) ? '"' : '\'';
    
    let res = [];
    
    res.push(quoteDelimiter);
    for (let c of s) {
      if (c == quoteDelimiter || c == '\\') {
        res.push('\\');
      }
      res.push(c);
    }
    res.push(quoteDelimiter);
    
    res = res.join("");
    return res;
  }
  
  toString() {
    let values = this._values.map( value => Tuple._escape(value)).toArray().join(", ");
    
    return "< " + values + " >";
  }
  
  *[Symbol.iterator](){
    for(let e of this._values.values()){
      yield e;
    }
  }
}