'use strict'

QUnit.module( "Schema" );


QUnit.test("Valid constructor calls", function( assert ) {
  let s;
  
  s = new Schema();
  assert.equal(s.size, 0, "no argument");
  
  s = new Schema([]);
  assert.equal(s.size, 0, "empty array");
  
  s = new Schema("a1");
  assert.equal(s.size, 1, "single string");

  s = new Schema("a1", "a2", "a3");
  assert.equal(s.size, 3, "several comma separated strings");

  s = new Schema(["a1", "a2", "a3"]);
  assert.equal(s.size, 3, "single arrayof strings");
  
  s = new Schema(["a1", "a2"], ["b1", "b2"]);
  assert.equal(s.size, 4, "several comma separated arrays of strings");
  
  s = new Schema(["a1", "a2"], ["b1", "b2"], []);
  assert.equal(s.size, 4, "several comma separated arrays of strings including an empty array");
  
  let s1 = new Schema("a1", "a2");
  let s2 = new Schema("b1", "b2", "b3");
  s = new Schema(s1);
  assert.equal(s.size, s1.size, "single schema");
  
  s = new Schema(s1, s2);
  assert.equal(s.size, s1.size + s2.size, "several schemas");
  
  s = new Schema([s1, s2]);
  assert.equal(s.size, s1.size + s2.size, "single array of schemas");
  
  s = new Schema(s1, [s2], "c1", "c2", ["d1", "d2", "d3", "d4", "d5"]);
  assert.equal(s.size, s1.size + s2.size + 2 + 5, "any combination of previous cases");
  
  s = new Schema([["a1", "a2"], [["b1", "b2"],["c1"]]]);
  assert.equal(s.size, 5, "nested arrays of strings");
  
  s = new Schema("   a1    ", "    a2", "a3");
  assert.equal(s.size, 3);
  assert.ok(s.hasAttribute("a1"));
  assert.ok(s.hasAttribute("a2"));
  assert.ok(s.hasAttribute("a3"));
  
});

QUnit.test("Duplicate attribute", function( assert ) {
  assert.throws(
    () => {
      new Schema("a1", "a3", "a1");
    },
    function(err) {
      console.log(err.toString());
      assert.ok(err.description.includes("duplicate attribute"));
      return err.type == AlgebraError.INVALID_ATTRIBUTE;
    },
    "duplicate attribute 'a1'"
  );
});
  
QUnit.test("Invalid constructor calls", function( assert ) {
  
  assert.throws(
    () => {
      new Schema(1);
    },
    function(err) {
      console.log(err.toString());
      return err.type == AlgebraError.INVALID_ATTRIBUTE;
    },
    "number argument"
  );

  assert.throws(
    () => {
      new Schema(null);
    },
    function(err) {
      console.log(err.toString());
      return err.type == AlgebraError.INVALID_ATTRIBUTE;
    },
    "null argument"
  );
  
  assert.throws(
    () => {
      new Schema(undefined);
    },
    function(err) {
      console.log(err.toString());
      return err.type == AlgebraError.INVALID_ATTRIBUTE;
    },
    "undefined argument"
  );
  
  assert.throws(
    () => {
      new Schema([1, 2, 3]);
    },
    function(err) {
      console.log(err.toString());
      return err.type == AlgebraError.INVALID_ATTRIBUTE;
    },
    "array of 'not-strings-or-schema'"
  );
  
  assert.throws(
    () => {
      new Schema("");
    },
    function(err) {
      console.log(err.toString());
      return err.type == AlgebraError.INVALID_ATTRIBUTE;
    },
    "empty string"
  );
  
  assert.throws(
    () => {
      new Schema("    ");
    },
    function(err) {
      console.log(err.toString());
      return err.type == AlgebraError.INVALID_ATTRIBUTE;
    },
    "string with only whitespaces"
  );
});

QUnit.test("Schema attribute setter", function(assert) {
  let s = new Schema(["a1", "a2", "a3", "a4"]);
  
  assert.throws(
    () => {
        s.attributes = null;
    },
    TypeError,
    "no setter for attributes"
  );
  
  let attrib = s.attributes.concat("Hello world");
  assert.equal(s.size, 4);
  
  console.log(attrib.toString());
});

QUnit.test("retrieve attribute index", function(assert) {
  let s = new Schema(["a1", "a2", "a3", "a4"]);
  
  assert.equal(s.getAttributeIndex("a1"), 0);
  assert.equal(s.getAttributeIndex("a2"), 1);
  assert.equal(s.getAttributeIndex("a3"), 2);
  assert.equal(s.getAttributeIndex("a4"), 3);
  assert.equal(s.getAttributeIndex("unknown"), undefined, "unknown value have undefined index");
});

QUnit.test("get attribute at index", function(assert) {
  let s = new Schema(["a1", "a2", "a3", "a4"]);
  
  assert.equal(s.getAttributeAt(0), "a1");
  assert.equal(s.getAttributeAt(1), "a2");
  assert.equal(s.getAttributeAt(2), "a3");
  assert.equal(s.getAttributeAt(3), "a4");
  
  assert.throws(
    () => {
        s.getAttributeAt(-1);
    },
    "out of bound index"
  );
  
  assert.throws(
    () => {
        s.getAttributeAt(4);
    },
    function(err) {
      console.log(err.toString());
      return err.type == AlgebraError.INVALID_INDEX;
    },
    "out of bound index"
  );
});

QUnit.test("check attribute presence", function(assert) {
  let s = new Schema(["a1", "a2", "a3", "a4"]);
  
  assert.ok(s.hasAttribute("a1"));
  assert.ok(s.hasAttribute("a2"));
  assert.ok(s.hasAttribute("a3"));
  assert.ok(s.hasAttribute("a4"));
  
  assert.notOk(s.hasAttribute());
  assert.notOk(s.hasAttribute("a5"));
  assert.notOk(s.hasAttribute(undefined));
});


QUnit.test("Union of disjoint schemas", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["b1", "b2", "b3"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs2);
  
  let union = s1.getUnion(s2);
  assert.ok(union instanceof Schema, "resulting union is a Schema");
  assert.equal(union.size, attribs1.length + attribs2.length);
});


QUnit.test("Union of schemas with non empty intersection", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["b1", "a2", "b3"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs2);
  
  let union = s1.getUnion(s2);
  assert.ok(union instanceof Schema, "resulting union is a Schema");
  assert.equal(union.size, 6);

  console.log(union.attributes.toString());
});

QUnit.test("Union of schema with itself", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  
  let s1 = new Schema(attribs1);
  
  let union = s1.getUnion(s1);
  assert.ok(union instanceof Schema, "resulting union is a Schema");
  assert.equal(union.size, attribs1.length);
  
  console.log(union.attributes.toString());
});

QUnit.test("Empty intersection", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["b1", "b2", "b3"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs2);
  
  let intersect = s1.getIntersection(s2);
  assert.ok(intersect instanceof Schema, "resulting intersection is a Schema");
  assert.equal(intersect.size, 0, "empty intersection");
});

QUnit.test("Not-empty intersection", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["b1", "b2", "a1", "a2"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs2);
  
  let intersect = s1.getIntersection(s2);
  assert.ok(intersect instanceof Schema, "resulting intersection is a Schema");
  assert.equal(intersect.size, 2, "not-empty intersection");
});

QUnit.test("Intersection with itself", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  
  let s1 = new Schema(attribs1);
  
  let intersect = s1.getIntersection(s1);
  assert.ok(intersect instanceof Schema, "resulting intersection is a Schema");
  assert.equal(intersect.size, s1.size, "not-empty intersection");
});

QUnit.test("Difference between disjoint schemas", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["b1", "b2", "b3"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs2);
  
  let diff = s1.getDifference(s2);
  assert.ok(diff instanceof Schema, "resulting intersection is a Schema");
  assert.equal(diff.size, s1.size);

  diff = s2.getDifference(s1);
  assert.ok(diff instanceof Schema, "resulting intersection is a Schema");
  assert.equal(diff.size, s2.size);
});

QUnit.test("Difference between intersecting schemas", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["b1", "b2", "b3", "a1", "a2"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs2);
  
  let diff = s1.getDifference(s2);
  assert.ok(diff instanceof Schema, "resulting intersection is a Schema");
  assert.equal(diff.size, 2);

  diff = s2.getDifference(s1);
  assert.ok(diff instanceof Schema, "resulting intersection is a Schema");
  assert.equal(diff.size, 3);
});

QUnit.test("Difference between identical schemas", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["a2", "a4", "a1", "a3"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs1);
  
  let diff = s1.getDifference(s1);
  assert.ok(diff instanceof Schema, "resulting difference is a Schema");
  assert.equal(diff.size, 0);

  diff = s1.getDifference(s2);
  assert.ok(diff instanceof Schema, "resulting difference is a Schema");
  assert.equal(diff.size, 0);

  diff = s2.getDifference(s1);
  assert.ok(diff instanceof Schema, "resulting difference is a Schema");
  assert.equal(diff.size, 0);
});


QUnit.test("Symmetric difference between disjoint schemas", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["b1", "b2", "b3"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs2);
  
  let diff = s1.getSymmetricDifference(s2);
  assert.ok(diff instanceof Schema, "resulting difference is a Schema");
  assert.equal(diff.size, s1.size + s2.size);

  diff = s2.getSymmetricDifference(s1);
  assert.ok(diff instanceof Schema, "resulting difference is a Schema");
  assert.equal(diff.size, s1.size + s2.size);
});


QUnit.test("Symmetric difference between identical schemas", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["a2", "a4", "a1", "a3"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs1);
  
  let diff = s1.getSymmetricDifference(s1);
  assert.ok(diff instanceof Schema, "resulting difference is a Schema");
  assert.equal(diff.size, 0);

  diff = s1.getSymmetricDifference(s2);
  assert.ok(diff instanceof Schema, "resulting difference is a Schema");
  assert.equal(diff.size, 0);

  diff = s2.getSymmetricDifference(s1);
  assert.ok(diff instanceof Schema, "resulting difference is a Schema");
  assert.equal(diff.size, 0);
});



QUnit.test("Loose equality", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["a2", "a4", "a1", "a3"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs1);

  assert.ok(s1.equals(s1), "schema equals to itself");
  
  assert.ok(s1.equals(s2), "attribute order is not revelant");
  assert.ok(s2.equals(s1), "equality is symmetric");
  
  let attribs3 = ["b1", "b2", "b3", "b4"];
  let s3 = new Schema(attribs3);
  
  assert.notOk(s1.equals(s3));
  assert.notOk(s3.equals(s1));
});


QUnit.test("Strict equality", function(assert) {
  let attribs1 = ["a1", "a2", "a3", "a4"];
  let attribs2 = ["a2", "a4", "a1", "a3"];
  
  let s1 = new Schema(attribs1);
  let s2 = new Schema(attribs1);

  assert.ok(s1.equalsStrict(s1));
  
  assert.ok(s1.equalsStrict(s2));
  assert.ok(s2.equalsStrict(s1));
  
  let attribs3 = ["a2", "a4", "a1", "a3"];
  let s3 = new Schema(attribs3);
  
  assert.notOk(s1.equalsStrict(s3));
  assert.notOk(s3.equalsStrict(s1));
  
  let attribs4 = ["b1", "b2", "b3", "b4"];
  let s4 = new Schema(attribs4);
  
  assert.notOk(s1.equalsStrict(s4));
  assert.notOk(s4.equalsStrict(s1));
});

QUnit.test("schema iterable on its attributes", function(assert) {
  
  let attribs1 = ["a1", "a2", "a3", "a4"];
  
  let s1 = new Schema(attribs1);
  
  let i = 0;
  for (let attrib of s1) {
    assert.equal(s1.getAttributeAt(i), attrib);
    i++;
  }
  
  assert.equal(i, s1.size);
  
});

QUnit.test("schema : corresponding attribute indexex from another schema", function(assert) {
  
  let indexes = 
      new Schema( ["a1", "a2", "a3", "a4"] )
          .getAttributeIndexes(
              new Schema( ["a4", "a2"] ) );

  assert.equal(indexes.length, 2);
  assert.equal(indexes[0], 3);
  assert.equal(indexes[1], 1);
});


QUnit.test("schema : to array", function(assert) {
  let attribs = ["a1", "a2", "a3", "a4"];
  
  let schema = new Schema(attribs);
  let ar = schema.toArray();
  console.log("to array :", ar);
  
  assert.ok(ar instanceof Array);
  assert.equal(ar.length, attribs.length);
  assert.ok(Immutable.is(Immutable.Seq(ar), Immutable.Seq(attribs)));
});



