'use strict'

QUnit.module( "Tuple" );

QUnit.test("constructor", function( assert ) {
  let t;
  
  t = new Tuple();
  assert.equal(t.size, 0, "empty tuple");
  
  t = new Tuple("val1");
  assert.equal(t.size, 1);
  
  t = new Tuple(["val1"]);
  assert.equal(t.size, 1);


  t = new Tuple("val1", "val2", "val3");
  assert.equal(t.size, 3);
  
  t = new Tuple(["val1", "val2", "val3"]);
  assert.equal(t.size, 3);
  

  t = new Tuple(["val1", "val2", "val3"]);
  let t2 = new Tuple(t);
  assert.equal(t2.size, 3);
  assert.equal(t2.size, t.size);
});



QUnit.test("set schema", function( assert ) {
  let t = new Tuple();
  let s = new Schema();
  
  t.schema = s;
  
  // schema can be set several times as long as it's not different
  t.schema = s;
  
  
  assert.equal(t.schema, s);
  assert.throws(
    ()=>{
      t.schema = new Schema();
    },
    function(err) {
      console.log(err.toString());
      return err.type == AlgebraError.INVALID_SCHEMA;
    },
    "schema can only be set once"
  );
  
  t = new Tuple();
  assert.throws(
    ()=>{
      t.schema = 123;
    },
    function(err) {
      console.log(err.toString());
      return err.type == AlgebraError.INVALID_SCHEMA;
    },
    "schema cannot only be set once"
  );
  
  t = new Tuple("val1", "val2", "val3");
  s = new Schema("a1", "a2", "a3");
  t.schema = s;
  
  
  t = new Tuple("val1", "val2", "val3");
  s = new Schema("a1", "a2");
  assert.throws(
    ()=>{
      t.schema = s;
    },
    "schema and tuple must have same size"
  );
});


QUnit.test("get value at index", function( assert ) {
  
  let values = ["val1", "val2", "val3", "val4"];
  let t = new Tuple(values);
  
  for(let i=0; i<values.length; i++){
    assert.equal(t.getValueAt(i), values[i]);
  }
  
  assert.throws(
    ()=>{
      t.getValueAt(-15);
    },
    "out of bound index"
  );

  assert.throws(
    ()=>{
      t.getValueAt(-5);
    },
    "out of bound index"
  );

  assert.throws(
    ()=>{
      t.getValueAt(1.5);
    },
    "index must be an integer"
  );

  assert.throws(
    ()=>{
      t.getValueAt();
    },
    "index must be an integer"
  );

  assert.throws(
    ()=>{
      t.getValueAt("2");
    },
    "index must be an integer"
  );
});


QUnit.test("get value of some attribute", function( assert ) {
  let values = ["val1", "val2", "val3", "val4"];
  let attribs = ["a1", "a2", "a3", "a4"];
  
  let t = new Tuple(values);
  let s = new Schema(attribs);
  
  t.schema = s;
  
  for (let i=0; i<values.length; i++){
    assert.equal(t.getValue(attribs[i]), values[i]);
  }
  
  assert.throws(
    ()=>{
      t.getValue();
    },
    "attrib name must be provided"
  );
  
  assert.throws(
    ()=>{
      t.getValue("unknown");
    },
    "attrib name must be part of schema"
  );
  
});


QUnit.test("as object", function( assert ) {

  let values = ["val1", "val2", "val3", "val4"];
  let attribs = ["a1", "a2", "a3", "a4"];
  
  let t = new Tuple(values);
  let s = new Schema(attribs);
  
  t.schema = s;
  
  let o = t.asObject();
  
  console.log(o);
  
  for (let i=0; i<values.length; i++){
    assert.equal(o[attribs[i]], values[i]);
  }
  


  t = new Tuple(values);
  o = t.asObject();
  console.log(o);
  
  for (let i=0; i<values.length; i++){
    assert.equal(o[i], values[i]);
  }
  

  t.schema = s;
  
  o = t.asObject();
  
  console.log(o);
  
  for (let i=0; i<values.length; i++){
    assert.equal(o[attribs[i]], values[i]);
  }
});


QUnit.test("tuple comparison", function( assert ) {
  let t1, t2;
  
  t1 = new Tuple();
  t2 = new Tuple();
  assert.equal(t1.compareTo(t2), 0);
  assert.equal(t2.compareTo(t1), 0);

  t1 = new Tuple("r");
  t2 = new Tuple("r");
  assert.equal(t1.compareTo(t2), 0);
  assert.equal(t2.compareTo(t1), 0);
  
  t1 = new Tuple(["a", "b", "c"]);
  t2 = new Tuple(["z", "b", "c"]);
  assert.ok(t1.compareTo(t2) < 0);
  assert.ok(t2.compareTo(t1) > 0);
  
  
  t1 = new Tuple(["a", "a", "a"]);
  t2 = new Tuple(["a", "a", "b"]);
  assert.ok(t1.compareTo(t2) < 0);
  assert.ok(t2.compareTo(t1) > 0);
  
  
  t1 = new Tuple(["a", "a", "a"]);
  t2 = new Tuple(["a"]);
  assert.throws(
    ()=>{
      t1.compareTo(t2);
    },
    "compared tuples must have same size"
  );
  
  assert.throws(
    ()=>{
      t1.compareTo(null);
    },
    "tuple must be compared to another tuple"
  );
  
  assert.throws(
    ()=>{
      t1.compareTo("hello");
    },
    "tuple must be compared to another tuple"
  );
  
  
  t1 = new Tuple(["b", "z"]);
  t2 = new Tuple(["y", "a"]);
  assert.ok(t1.compareTo(t2) < 0, "b < y");
  assert.ok(t2.compareTo(t1) > 0, "y > b");
  
  // when both tuple have a schema, comparison is made according sorted attributes
  
  t1 = new Tuple(["b", "z"]);
  t1.schema = new Schema("A1", "A2");
  t2 = new Tuple(["y", "a"]);
  t2.schema = new Schema("A2", "A1");
  assert.ok(t1.compareTo(t2) > 0, "t1:A1:b > t2:A1:a");
  assert.ok(t2.compareTo(t1) < 0, "t2:A1:a < t1:A1:b");
  
  t1 = new Tuple(["a", "z"]);
  t1.schema = new Schema("A1", "A2");
  t2 = new Tuple(["y", "b"]);
  t2.schema = new Schema("A2", "A1");
  assert.ok(t1.compareTo(t2) < 0, "t1:A1:a < t2:A1:b");
  assert.ok(t2.compareTo(t1) > 0, "t2:A1:b > t1:A1:z");
});


QUnit.test("tuple equality", function( assert ) {
  let values = ["val1", "val2", "val3", "val4"];
  let attribs = ["a1", "a2", "a3", "a4"];
  let schema = new Schema(attribs);
  
  let t1 = new Tuple(values);
  let t2 = new Tuple(values);
  assert.ok(t1.equals(t2), "two tuples without schema");
  assert.ok(t2.equals(t1));
  
  t1 = new Tuple(values);
  t1.schema = schema;
  t2 = new Tuple(values);
  assert.ok(t1.equals(t2), "one tuple with schema, other tuple has no schema");
  assert.ok(t2.equals(t1));
  
  t1 = new Tuple(["val1", "val2", "val3"]);
  t1.schema = new Schema("a1", "a2", "a3")
  t2 = new Tuple(["val3", "val1", "val2"]);
  t2.schema = new Schema("a3", "a1", "a2")
  assert.ok(t1.equals(t2), "same schema but different order");
  assert.ok(t2.equals(t1));
  
  
  t1 = new Tuple(["val1", "val2", "val3"]);
  t1.schema = new Schema("a1", "a2", "a3")
  t2 = new Tuple(["val1", "val2", "val3"]);
  t2.schema = new Schema("a3", "a1", "a2")
  assert.notOk(t1.equals(t2), "values are the same, but different attribute order");
  assert.notOk(t2.equals(t1));
  

  t1 = new Tuple(["val1", "val2", "val3"]);
  t1.schema = new Schema("a1", "a2", "a3")
  t2 = new Tuple(["val1", "val2", "val3"]);
  t2.schema = new Schema("b1", "b2", "b3")
  assert.notOk(t1.equals(t2), "values are the same, but different schema");
  assert.notOk(t2.equals(t1));
  
  
  t1 = new Tuple();
  t2 = new Tuple();
  assert.ok(t1.equals(t2), "tuple with no attribute");
  assert.ok(t2.equals(t1));
  
  
});

