'use strict'


Parser.load('./parser.peg', parser => {
  
  QUnit.module( "Attribute values" );
  
  

  QUnit.test("unquoted values with no space", function( assert ) {
    let source = `@let r = @relation[a1, a2, a3]{
        <la première valeur,la deuxième valeur,la troisième valeur>
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 3);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValueAt(0), "la première valeur");
    assert.equal(t.getValueAt(1), "la deuxième valeur");
    assert.equal(t.getValueAt(2), "la troisième valeur");
  });

  QUnit.test("unquoted values with regular space after comma", function( assert ) {
    let source = `@let r = @relation[a1, a2, a3]{
        < la première valeur, la deuxième valeur, la troisième valeur >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 3);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValueAt(0), "la première valeur");
    assert.equal(t.getValueAt(1), "la deuxième valeur");
    assert.equal(t.getValueAt(2), "la troisième valeur");
  });

  QUnit.test("unquoted values with extra spaces inbetween", function( assert ) {
    let source = `@let r = @relation[a1, a2, a3]{
        <       la première valeur        ,            la deuxième valeur      ,  la troisième valeur     >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 3);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValueAt(0), "la première valeur");
    assert.equal(t.getValueAt(1), "la deuxième valeur");
    assert.equal(t.getValueAt(2), "la troisième valeur");
  });


  QUnit.test("unquoted values with escaped chars", function( assert ) {
    let source = `@let r = @relation[a1]{
        < c:\\windows >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);

    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `c:\\windows`);
  });

  QUnit.test("unquoted values with escaped chars", function( assert ) {
    let source = `@let r = @relation[a1]{
        < c:\\\\java >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `c:\\java`);
  });
  

  QUnit.test("unquoted values with escaped chars", function( assert ) {
    let source = `@let r = @relation[a1]{
        < Lundi\\, j'ai répondu : "non\\, non !" > 
        
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `Lundi, j'ai répondu : "non, non !"`);
  });

  QUnit.test("just one single-quoted value with no escape char", function( assert ) {
    let source = `@let r = @relation[a1]{
        <'une valeur' >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "une valeur");
  });
  
  
  
  QUnit.test("single-quoted values with no escaped characters", function( assert ) {
    let source = `@let r = @relation[a1, a2, a3]{
        <'la première valeur','la deuxième valeur','la troisième valeur'>
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 3);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValueAt(0), "la première valeur");
    assert.equal(t.getValueAt(1), "la deuxième valeur");
    assert.equal(t.getValueAt(2), "la troisième valeur");
  });
  
  QUnit.test("single-quoted values with standard space between values", function( assert ) {
    let source = `@let r = @relation[a1, a2, a3]{
        < 'la première valeur', 'la deuxième valeur', 'la troisième valeur' >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 3);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValueAt(0), "la première valeur");
    assert.equal(t.getValueAt(1), "la deuxième valeur");
    assert.equal(t.getValueAt(2), "la troisième valeur");
  });
  
  
  QUnit.test("single-quoted values with escaped chars", function( assert ) {
    let source = `@let r = @relation[a1]{
        < 'c:\\windows' >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);

    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `c:\\windows`);
  });

  QUnit.test("single-quoted values with escaped chars", function( assert ) {
    let source = `@let r = @relation[a1]{
        < 'c:\\\\java' >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `c:\\java`);
  });
  
  QUnit.test("double-quoted values with backslash not escaping next char", function( assert ) {
    let source = `@let r = @relation[a1]{
        < "c:\\windows" >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);

    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `c:\\windows`);
  });

  QUnit.test("double-quoted values with escaped backslash", function( assert ) {
    let source = `@let r = @relation[a1]{
        < "c:\\\\java" >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `c:\\java`);
  });
  



  QUnit.test("double-quoted values with escaped double quote", function( assert ) {
    let source = `@let r = @relation[a1]{
        < "Mardi, j'ai répondu : \\"non, non !\\"" >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `Mardi, j'ai répondu : "non, non !"`);
  });
  

  QUnit.test("single-quoted value with escaped single quote", function( assert ) {
    let source = `@let r = @relation[a1]{
        < 'Jeudi, j\\'ai répondu : "non, non !"' >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `Jeudi, j'ai répondu : "non, non !"`);
  });
  

  QUnit.test("double-quoted value ending with backslash", function( assert ) {
    let source = `@let r = @relation[a1]{
        < "une chaîne qui se termine par un backslash \\\\" >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `une chaîne qui se termine par un backslash \\`);
  });
  


  QUnit.test("unquoted value ending with backslash", function( assert ) {
    let source = `@let r = @relation[a1]{
        < une autre chaîne qui se termine par un backslash \\ >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), `une autre chaîne qui se termine par un backslash \\`);
  });
  
  
  QUnit.test("single quoted empty value", function( assert ) {
    let source = `@let r = @relation[a1]{
        < '' >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "");
  });
  
  
  QUnit.test("double quoted empty value", function( assert ) {
    let source = `@let r = @relation[a1]{
        < "" >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "");
  });
  
  

  QUnit.test("single quoted value with only spaces", function( assert ) {
    let source = `@let r = @relation[a1]{
        < '     ' >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "     ");
  });
  
  QUnit.test("double quoted value with only spaces", function( assert ) {
    let source = `@let r = @relation[a1]{
        < "     " >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "     ");
  });
  
  QUnit.test("single quoted value with leading spaces", function( assert ) {
    let source = `@let r = @relation[a1]{
        < '     la valeur' >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "     la valeur");
  });
  
  QUnit.test("double quoted value with leading spaces", function( assert ) {
    let source = `@let r = @relation[a1]{
        < "     la valeur" >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "     la valeur");
  });
  

  
  QUnit.test("single quoted value with trailing spaces", function( assert ) {
    let source = `@let r = @relation[a1]{
        < 'la valeur     ' >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "la valeur     ");
  });
  
  QUnit.test("double quoted value with trailing spaces", function( assert ) {
    let source = `@let r = @relation[a1]{
        < "la valeur     " >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "la valeur     ");
  });


  QUnit.test("single quoted value with leading and trailing spaces", function( assert ) {
    let source = `@let r = @relation[a1]{
        < '     la valeur     ' >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "     la valeur     ");
  });
  
  QUnit.test("double quoted value with leading and trailing spaces", function( assert ) {
    let source = `@let r = @relation[a1]{
        < "     la valeur     " >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "     la valeur     ");
  });


  QUnit.test("single quoted value with missing closing single quote", function( assert ) {
    let source = `@let r = @relation[a1]{
        < 'la valeur      >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "'la valeur");
  });

  QUnit.test("double quoted value with missing closing double quote", function( assert ) {
    let source = `@let r = @relation[a1]{
        < "la valeur      >
    }`;
    console.log("source: ´" + source + "´");
    
    let {database:db} = parser.parse(source);
    
    console.log(db.toString());
    
    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 1);
    assert.equal(r.tupleSet.size, 1);
    
    let t = r.tupleSet.getTupleAt(0);
    assert.equal(t.getValue("a1"), "\"la valeur");
  });

  
  QUnit.test("multi-line unquoted value", function( assert ) {
    let source = `@let r = @relation[a1]{
        < la valeur      
        est sur deux lignes >
    }`;
    console.log("source: ´" + source + "´");
    
    assert.throws(
      () => {parser.parse(source)},
      function(err) {
        console.log(err.toString());
        assert.equal(err.type, SqfError.SYNTAX);
        assert.equal(err.location.start.line, 3, "value cannot continue on third line");
        assert.ok(err.description.includes("Expected"));
        return true;
      },
      "value cannot span multiple lines"
    );
  });

  QUnit.test("multi-line single quoted value", function( assert ) {
    let source = `@let r = @relation[a1]{
        < 'la valeur      
        est sur deux lignes'
        >
    }`;
    console.log("source: ´" + source + "´");
    
    assert.throws(
      () => {
        parser.parse(source)
      },
      function(err) {
        console.log(err.toString());
        assert.equal(err.type, SqfError.SYNTAX);
        return true;
      },
      "value cannot span multiple lines"
    );
  });

  
});





