'use strict'


Parser.load('./parser.peg', parser => {
  
  QUnit.module( "Expression" );
  
  QUnit.test("relation", function( assert ) {
    let source = `@let r = @relation[prénom, nom] {
      < Isabelle    , Dello Strito >
      < Sergio      , Mazzoleni    >
      < Christophe  , Raes         >
    }`
    
    let {database:db} = parser.parse(source);

    assert.equal(db.size, 1);
    let r = db.get("r");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 2);
    assert.equal(r.tupleSet.size, 3);
    assert.equal(r.toExpressionString(), "r");
  });

  QUnit.test("id", function( assert ) {
    let source = 
      ` @let r = @relation[prénom, nom] {
          < Isabelle    , Dello Strito >
          < Sergio      , Mazzoleni    >
          < Christophe  , Raes         >
        }
        
        @let alias = r
        `
        
    let {database:db} = parser.parse(source);

    assert.equal(db.size, 2);
    let r = db.get("r");
    let alias = db.get("alias");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 2);
    assert.equal(r.tupleSet.size, 3);
    
    assert.ok(alias != null);
    assert.equal(alias.schema, r.schema);
    assert.ok(alias.tupleSet.equals(r.tupleSet));

    assert.equal(alias.toExpressionString(), "alias");
    
    console.log(alias.toString());
  });
  
  QUnit.test("select", function( assert ) {
    let source = 
      ` @let r = @relation[prénom, nom] {
          < Isabelle    , Dello Strito >
          < Sergio      , Mazzoleni    >
          < Franca      , Mazzoleni    >
          < Eugenio     , Mazzoleni    >
          < Christophe  , Raes         >
        }
        
        @let s = @select { nom = 'Raes'} r

        @let mazzo = @select { nom = "Mazzoleni"} r
        `
        
    let {database:db} = parser.parse(source);

    assert.equal(db.size, 3);
    let r = db.get("r");
    let s = db.get("s");
    let mazzo = db.get("mazzo");
    
    console.log(r.toExpressionString());
    console.log(s.toExpressionString());
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 2);
    assert.equal(r.tupleSet.size, 5);
    
    assert.ok(s != null);
    console.log(s.toString());
    assert.ok(s.schema.equals(r.schema), "selection has same schema as expression operand");
    assert.equal(s.tupleSet.size, 1, "one tuple whose name is 'Raes'");

    console.log(mazzo.toString());
    assert.equal(mazzo.tupleSet.size, 3, "3 mazzoleni entries in r");
    
    assert.equal(s.toExpressionString(), "@select{nom='Raes'} r");
    assert.equal(mazzo.toExpressionString(), "@select{nom='Mazzoleni'} r");
  });
  
  QUnit.test("projection", function( assert ) {
    let source = 
      ` @let r = @relation[prénom, nom] {
          < Isabelle    , Dello Strito >
          < Sergio      , Mazzoleni    >
          < Christophe  , Ronck        >
          < Franca      , Mazzoleni    >
          < Eugenio     , Mazzoleni    >
          < Christophe  , Raes         >
        }
        
        @let noms = @project {nom} r
        @let prénoms = @project {prénom} r
`
        
    let {database:db} = parser.parse(source);
    
    let r = db.get("r");
    let noms = db.get("noms");
    let prénoms = db.get("prénoms");
    
    console.log(noms.toExpressionString());
    console.log(prénoms.toExpressionString());
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 2);
    assert.equal(r.tupleSet.size, 6);
    
    assert.ok(noms != null);
    assert.ok(prénoms != null);
    
    console.log(noms.toString());
    console.log(prénoms.toString());
    
    assert.equal(noms.schema.size, 1);
    assert.equal(prénoms.schema.size, 1);
    
    
    assert.equal(noms.tupleSet.size, 4)
    assert.equal(prénoms.tupleSet.size, 5)
    
    assert.equal(noms.toExpressionString(), "@project{nom} r");
    assert.equal(prénoms.toExpressionString(), "@project{prénom} r");
    
    
  });
  
  QUnit.test("join", function( assert ) {
    let source = 
      ` @let hommes = @relation[homme] {
          < Sergio      >
          < Christophe  >
          < Pascal      >
        }
        
        @let femmes = @relation[femme] {
          < Isabelle    >
          < Caroline  >
        }
        
        @let couples = hommes * femmes
`
        
    let {database:db} = parser.parse(source);
    
    let couples = db.get("couples");
    console.log(couples.toString());
    
    assert.equal(couples.schema.size, 2);
    assert.ok(couples.schema.hasAttribute("homme"));
    assert.ok(couples.schema.hasAttribute("femme"));
    assert.equal(couples.tupleSet.size, 6);
    
    console.log(couples.toExpressionString());
    assert.equal(couples.toExpressionString(), "hommes * femmes");

  });
  
  QUnit.test("rename", function( assert ) {
    let source = 
      ` @let r = @relation[prénom, nom] {
          < Isabelle    , Dello Strito >
          < Sergio      , Mazzoleni    >
          < Christophe  , Ronck        >
          < Franca      , Mazzoleni    >
          < Eugenio     , Mazzoleni    >
          < Christophe  , Raes         >
        }
        
        @let e = @rename{nom:lastname, prénom:firstname} r
`
        
    let {database:db} = parser.parse(source);
    let r = db.get("r");
    let e = db.get("e");
    console.log(e.toString());
    
    assert.equal(e.schema.size, 2);
    assert.ok(e.schema.hasAttribute("lastname"));
    assert.ok(e.schema.hasAttribute("firstname"));
    assert.equal(e.tupleSet.size, r.tupleSet.size);
    
    for (let t of r.tupleSet) {
      let copy = new Tuple(t);
      assert.ok(e.tupleSet.contains(copy));
    }
    
    console.log(e.toExpressionString());
    assert.equal(e.toExpressionString(), "@rename{nom:lastname, prénom:firstname} r");
  });
  
  QUnit.test("rename no param", function( assert ) {
    let source = 
      ` @let r = @relation[prénom, nom] {
          < Isabelle    , Dello Strito >
          < Sergio      , Mazzoleni    >
          < Christophe  , Ronck        >
          < Franca      , Mazzoleni    >
          < Eugenio     , Mazzoleni    >
          < Christophe  , Raes         >
        }
        
        @let e = @rename{} r
`
        
    let {database:db} = parser.parse(source);
    let r = db.get("r");
    let e = db.get("e");
    console.log(e.toString());
    
    assert.equal(e.schema.size, 2);
    assert.equal(e.tupleSet.size, r.tupleSet.size);
    
    for (let t of r.tupleSet) {
      let copy = new Tuple(t);
      assert.ok(e.tupleSet.contains(copy));
    }
    
    console.log(e.toExpressionString());
    assert.equal(e.toExpressionString(), "@rename{} r");
  });
  

  QUnit.test("Union", function( assert ) {
    let source = 
      ` @let r1 = @relation[prénom, nom] {
          < Sergio      , Mazzoleni    >
          < Christophe  , Ronck        >
          < Christophe  , Raes         >
        }
        
        @let r2 = @relation[prénom, nom] {
          < Isabelle    , Dello Strito >
          < Franca      , Mazzoleni    >
          < Eugenio     , Mazzoleni    >
          < Christophe  , Raes         >
        }
        
        @let e = r1 + r2
`

    let {database:db} = parser.parse(source);
    let e = db.get("e");
    
    console.log(e.toString());
    assert.equal(e.tupleSet.size, 6);
    console.log(e.toExpressionString());
    
    assert.equal(e.toExpressionString(), "r1 + r2");
  });
  
  QUnit.test("Difference", function( assert ) {
    let source = 
      ` @let r1 = @relation[prénom, nom] {
          < Sergio      , Mazzoleni    >
          < Christophe  , Ronck        >
          < Christophe  , Raes         >
        }
        
        @let r2 = @relation[prénom, nom] {
          < Isabelle    , Dello Strito >
          < Franca      , Mazzoleni    >
          < Eugenio     , Mazzoleni    >
          < Christophe  , Raes         >
        }
        
        @let e = r1 - r2
`

    let {database:db} = parser.parse(source);
    let e = db.get("e");
    
    console.log(e.toString());
    assert.equal(e.tupleSet.size, 2);
    console.log(e.toExpressionString());
    assert.equal(e.toExpressionString(), "r1 - r2");
  });
  
  QUnit.test("Parentheses", function( assert ) {
    let source = 
      ` @let r = @relation[prénom, nom] {
          < Isabelle    , Dello Strito >
          < Sergio      , Mazzoleni    >
          < Christophe  , Raes         >
        }
        
        @let alias = (r)
        `
        
    let {database:db} = parser.parse(source);

    assert.equal(db.size, 2);
    let r = db.get("r");
    let alias = db.get("alias");
    
    assert.ok(r != null);
    assert.equal(r.schema.size, 2);
    assert.equal(r.tupleSet.size, 3);
    
    assert.ok(alias != null);
    assert.equal(alias.schema, r.schema);
    assert.ok(alias.tupleSet.equals(r.tupleSet));
    
    console.log(alias.toString());
    assert.equal(alias.toExpressionString(), "alias");
  });
});





